# flake8: noqa
from gluepy.conf import default_settings
from gluepy.utils.loading import LazyProxy, import_string
from .base import BaseDataManager
from .pandas import PandasDataManager


data_manager: BaseDataManager = LazyProxy(
    lambda: import_string(default_settings.DATA_BACKEND)()
)



import yaml
import logging
from io import StringIO
from gluepy.conf import default_context
from gluepy.files.storages import default_storage

logger = logging.getLogger(__name__)

REGISTRY = {}


class Task:
    label = None

    def __init__(self) -> None:
        self.label == self.label or self.__class__.__name__.lower()

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        REGISTRY[cls.label] = cls

    def run(self):
        raise NotImplementedError()


class BootstrapTask(Task):
    """Used by default as the START_TASK and injected in each DAG.

    Provide various bootstrapping functionality such as serializing context
    at start of execution.

    """

    label = "bootstraptask"

    def run(self):
        logger.debug(
            f"""
            Run ID: {default_context.gluepy.run_id}
            Run Folder: {default_context.gluepy.run_folder}
            """
        )
        default_storage.touch(
            default_storage.runpath("context.yaml"),
            StringIO(yaml.dump(default_context.to_dict())),
        )



import sys
from argparse import ArgumentParser
from gluepy import VERSION
from gluepy.exceptions import CommandError


REGISTRY = {}


class Command(object):
    """Define a command that will be exposed to the CLI"""

    label = None
    parser_class = ArgumentParser
    autoload = True

    def __init_subclass__(cls, **kwargs):
        """Register the class in a REGISTRY to be used from cli"""
        super().__init_subclass__(**kwargs)
        if not cls.autoload:
            return

        name = cls.label or cls.__name__.lower()
        REGISTRY[name] = cls

    def handle(self, *args, **kwargs):
        """Hook to add logic to your command"""
        raise NotImplementedError("Command has not been implemented yet.")

    def add_arguments(self, parser):
        """Hook to add additional arguments to the parser"""
        pass

    def stdout(self, message):
        print(message)

    def stderr(self, message):
        print(message)

    def run(self):
        parser = self.parser_class()
        self.add_arguments(parser)
        kwargs = vars(parser.parse_args(args=None if sys.argv[1:] else ["--help"]))
        self.handle(**kwargs)


class DefaultCommand(Command):
    """Default command used when no command is provided.

    e.g. `.run.py --version` will use this command.

    """

    autoload = False

    def add_arguments(self, parser):
        parser.add_argument(
            "--list", "-l", help="List all available commands", action="store_true"
        )
        parser.add_argument(
            "--version", help="Print version of Gluepy", action="store_true"
        )

    def handle(self, **options):
        if not options:
            self.parser_class().print_help()
        elif options.get("version"):
            self.stdout("Gluepy version: %s" % VERSION)
        elif options.get("list"):
            self.stdout(
                "Available commands: \n%s" % "\n".join([cmd for cmd in REGISTRY.keys()])
            )


def call_command():
    """Main method that parse which command to call and calls it"""
    if len(sys.argv) <= 1 or sys.argv[1].startswith("-"):
        DefaultCommand().run()
        return

    cmd_label = sys.argv.pop(1)

    try:
        cmd = REGISTRY[cmd_label]()
    except KeyError:
        raise CommandError(f"The command '{cmd_label}' is not defined.")

    cmd.run()



import os
from gluepy.utils.loading import LazyProxy, import_module, SingletonMixin


class Settings(SingletonMixin):
    def __init__(self, dotted_path: str, *args, **kwargs):
        module = import_module(dotted_path)
        for key in dir(module):
            if key.isupper():
                setattr(self, key, getattr(module, key))


default_settings = LazyProxy(lambda: Settings(os.environ.get("GLUEPY_SETTINGS_MODULE")))



import os
import logging
from pathlib import Path
from typing import Union
from io import StringIO, BytesIO
from google.cloud import storage
from google.api_core.exceptions import NotFound as GoogleNotFound
from gluepy.conf import default_settings
from .base import BaseStorage

logger = logging.getLogger(__name__)


class GoogleStorage(BaseStorage):
    """ "
    Storage support for Google Cloud Storage.
    """

    separator = "/"

    def __init__(self) -> None:
        super().__init__()
        self.client = storage.Client()
        self.bucket_name = getattr(default_settings, "GOOGLE_GCS_BUCKET", None)
        assert self.bucket_name, "GOOGLE_GCS_BUCKET setting must be set."
        self.bucket = self.client.bucket(self.bucket_name)
        self.storage_root = default_settings.STORAGE_ROOT.lstrip(self.separator)

    def touch(self, file_path: str, content: Union[StringIO, BytesIO]) -> None:
        """Create a new blob at file path.

        Args:
            file_path (str): Path to file we want to create
            content (Union[StringIO, BytesIO]): Content of file we want to generate

        Raises:
            NotImplementedError: Base class raise NotImplementedError
        """
        assert not file_path.endswith(
            self.separator
        ), f"File name cannot end with '{self.separator}'."
        content.seek(0)
        if isinstance(content, StringIO):
            content = BytesIO(content.read().encode("utf-8"))
        self.bucket.blob(self.abspath(file_path)).upload_from_file(content)

    def open(self, file_path: str, mode: str = "rb") -> Union[str, bytes]:
        """Opens a blob at file_path

        Args:
            file_path (str): File path of blob we want to open

        Raises:
            NotImplementedError: Base class raise NotImplementedError
        """
        if mode != "rb":
            logger.warning(f"`mode` is not used for {self.__class__.__name__}")

        stream = BytesIO()
        try:
            self.bucket.blob(self.abspath(file_path)).download_to_file(stream)
        except GoogleNotFound as exc:
            raise FileNotFoundError(
                f"File '{self.abspath(file_path)}' does not exist."
            ) from exc
        stream.seek(0)
        return stream.read()

    def rm(self, path: str, recursive: bool = False) -> None:
        """Delete a file

        Args:
            path (str): Path to file to delete
            recursive (bool): If allowed to delete recursive directories or not.

        Raises:
            NotImplementedError: Base class raise NotImplementedError
        """
        if self.isdir(path):
            files, dirs = self.ls(path)
            if (files or dirs) and not recursive:
                raise FileExistsError(
                    "Option `recursive` must be set to delete directories that "
                    "contain other directories or files."
                )
            for item in files + dirs:
                self.rm(item, recursive=recursive)
        else:
            self.bucket.blob(self.abspath(path=path)).delete()

    def cp(
        self,
        src_path: str,
        dest_path: str,
        recursive: bool = False,
        overwrite: bool = False,
    ) -> None:
        """Copy a file from source to destination

        Args:
            src_path (str): Path to file or directory to copy.
            dest_path (str): Path to file or directory to copy to.
            recursive (bool): If should copy sub directories as well.
            overwrite (bool): If should copy to destination that already exists.
        """
        if not self.exists(src_path):
            raise FileNotFoundError(f"File at '{src_path}' not found.")
        if not self.exists(dest_path) and not overwrite:
            raise FileNotFoundError(f"File at '{dest_path}' not found.")

        if self.isfile(src_path):
            self.touch(dest_path, self.open(src_path))
        elif self.isdir():
            files, dirs = self.ls(src_path)
            if dirs and not recursive:
                raise FileExistsError(
                    "Option `recursive` must be set to copy directories that "
                    "contain other directories."
                )
            paths = files + dirs
            for path in paths:
                self.cp(
                    path,
                    str(
                        Path(dest_path.lstrip(self.separator))
                        / str(Path(path))
                        .relative_to(src_path.lstrip(self.separator))
                        .lstrip(self.separator)
                    ),
                    recursive=recursive,
                    overwrite=overwrite,
                )
        else:
            raise ValueError(
                f"Cannot identify if path '{src_path}' is a file or directory."
            )

    def ls(self, path: str) -> tuple[list[str], list[str]]:
        """List all files and directories at given path.

        Args:
            path (str): Path where we want to list contents of

        Raises:
            NotImplementedError: Base class raise NotImplementedError

        Returns:
            Tuple[List[str], List[str]]:
              First list is files, second list is directories.
        """
        files: list[str] = []
        dirs: list[str] = []

        for blob in self.client.list_blobs(self.bucket, prefix=self.abspath(path)):
            if self.isdir(blob.name):
                dirs += (
                    [
                        f"{self.relpath(blob.name).rstrip(self.separator)}{self.separator}"  # noqa
                    ]
                    if self._is_in_root(blob.name, path)
                    else []
                )
            else:
                if self._is_in_root(os.path.dirname(blob.name), path) and (
                    os.path.dirname(blob.name) != os.path.dirname(self.abspath(path))
                ):
                    dirs += [
                        f"{self.relpath(os.path.dirname(blob.name)).rstrip(self.separator)}{self.separator}"  # noqa
                    ]
                files += (
                    [self.relpath(blob.name)]
                    if self._is_in_root(blob.name, path)
                    else []
                )

        return (
            files,
            dirs,
        )

    def mkdir(self, path: str, make_parents: bool = False) -> None:
        """Make a new directory at location

        Args:
            path (str): Path of directory we want to create
            make_parents (bool, optional): If we should generate parents
              folders as well. Defaults to False.

        Raises:
            NotImplementedError: Base class raise NotImplementedError
        """
        path = path.rstrip(self.separator) + self.separator
        if self.exists(path):
            logger.warning("Directory '%s' already exists.", path)
            return

        parent = os.path.dirname(path)
        if not self.exists(parent) and not make_parents:
            raise FileNotFoundError(
                f"Parent directory '{parent}' does not exist. "
                "Use option `make_parents` to automatically create parent directories."
            )

        self.bucket.blob(self.abspath(path)).upload_from_file(BytesIO())

    def isdir(self, path: str) -> bool:
        """Check if path is directory or not.

        Args:
            path (str): Path we want to check

        Raises:
            NotImplementedError: Base class raise NotImplementedError

        Returns:
            bool: True/False if path is directory or not
        """
        # Ensure always ending with `/`.
        path = self.abspath(path).rstrip(self.separator) + self.separator
        return self.bucket.blob(path).exists() or bool(
            list(self.client.list_blobs(self.bucket, prefix=path, max_results=1))
        )

    def isfile(self, path: str) -> bool:
        """Check if path is a file or not.

        Args:
            path (str): Path we want to check

        Raises:
            NotImplementedError: Base class raise NotImplementedError

        Returns:
            bool: True/False if path is file or not.
        """
        return self.bucket.blob(self.abspath(path=path)).exists() and not path.endswith(
            self.separator
        )

    def exists(self, path: str) -> bool:
        """Check if path exists or not.

        Args:
            path (str): Path we want to check

        Raises:
            NotImplementedError: Base class raise NotImplementedError

        Returns:
            bool: True/False if path is file or not.
        """
        return self.isfile(path) or self.isdir(path)

    def _is_in_root(self, blob_name: str, dir_path: str) -> bool:
        return os.path.dirname(blob_name.strip(self.separator)) == self.abspath(
            dir_path
        ).strip(self.separator)



from typing import List

from gluepy.exec.tasks import Task
from gluepy.utils.loading import import_string
from gluepy.conf import default_settings


class DAG:
    label = None
    extra_options = {}
    tasks = []

    def __init__(self) -> None:
        self.label == self.label or self.__class__.__name__.lower()

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        REGISTRY[cls.label] = cls

    def inject_tasks(self) -> List[Task]:
        """Inject all tasks including START_TASK to the final
        list of executable tasks of this DAG.

        Returns:
            List[Task]: Full list of tasks of DAG.
        """
        return [import_string(default_settings.START_TASK)] + self.tasks


REGISTRY = {}



import os
import logging
from typing import List, Tuple, Union
from pathlib import Path
from io import StringIO, BytesIO
import boto3
from botocore.client import Config
from botocore.exceptions import ClientError
from gluepy.conf import default_settings
from gluepy.files.storages.base import BaseStorage

logger = logging.getLogger(__name__)


class S3Storage(BaseStorage):
    """
    Storage backend to support S3 based storages such as
    AWS S3 or DigitalOcean Spaces.
    """

    MAX_CHUNK_SIZE = 32_000_000

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._connection = None
        self._bucket = None
        self.region_name = default_settings.AWS_S3_REGION_NAME
        self.use_ssl = True
        self.endpoint_url = default_settings.AWS_S3_ENDPOINT_URL
        self.config = Config(
            s3={"addressing_style": None},
            signature_version=None,
            proxies=None,
        )
        self.verify = True
        self.access_key, self.secret_key = (
            default_settings.AWS_ACCESS_KEY_ID,
            default_settings.AWS_SECRET_ACCESS_KEY,
        )

    @property
    def connection(self):
        if self._connection is None:
            session = boto3.session.Session(
                aws_access_key_id=self.access_key,
                aws_secret_access_key=self.secret_key,
            )
            self._connection = session.resource(
                "s3",
                region_name=self.region_name,
                use_ssl=self.use_ssl,
                endpoint_url=self.endpoint_url,
                config=self.config,
                verify=self.verify,
            )

        return self._connection

    @property
    def bucket(self):
        if self._bucket is None:
            self._bucket = self.connection.Bucket(
                default_settings.AWS_STORAGE_BUCKET_NAME
            )
        return self._bucket

    def touch(self, file_path: str, content: Union[StringIO, BytesIO]) -> None:
        """Create a new blob at file path.

        Args:
            file_path (str): Path to file we want to create
            content (Union[StringIO, BytesIO]): Content of file we want to generate

        Raises:
            NotImplementedError: Base class raise NotImplementedError
        """
        if isinstance(content, StringIO):
            content = self._to_bytes(content)
        obj = self.bucket.Object(file_path)
        content.seek(0, os.SEEK_SET)
        obj.upload_fileobj(content)
        return file_path

    def open(self, file_path: str, mode: str = "rb") -> Union[str, bytes]:
        """Opens a blob at file_path

        Args:
            file_path (str): File path of blob we want to open

        Raises:
            NotImplementedError: Base class raise NotImplementedError
        """
        logger.debug(f"Reading file from path '{file_path}'")
        obj = self.bucket.Object(file_path)
        handler = BytesIO()
        obj.download_fileobj(handler)
        handler.seek(0, os.SEEK_SET)
        return (
            handler.read()
            if "b" in mode
            else handler.read().decode("utf-8", errors="ignore")
        )

    def ls(self, path: str) -> Tuple[List[str], List[str]]:
        """List all files and directories at given path.

        Args:
            path (str): Path where we want to list contents of

        Raises:
            NotImplementedError: Base class raise NotImplementedError

        Returns:
            Tuple[List[str], List[str]]:
              First list is files, second list is directories.
        """
        path = self.abspath(path)
        # The path needs to end with a slash, but if the root is empty, leave it.
        if path and not path.endswith("/"):
            path += "/"

        directories = []
        files = []
        paginator = self.connection.meta.client.get_paginator("list_objects")
        pages = paginator.paginate(
            Bucket=default_settings.AWS_STORAGE_BUCKET_NAME,
            Delimiter="/",
            Prefix=path,
            PaginationConfig={"PageSize": 1000},
        )
        for page in pages:
            for entry in page.get("CommonPrefixes", ()):
                directories.append(self.relpath(entry["Prefix"]))
            for entry in page.get("Contents", ()):
                key = entry["Key"]
                if key != path:
                    files.append(self.relpath(key))
        return directories, files

    def mkdir(self, path: str, make_parents: bool = False) -> None:
        """Make a new directory at location

        Args:
            path (str): Path of directory we want to create
            make_parents (bool, optional): If we should generate parents
              folders as well. Defaults to False.

        Raises:
            NotImplementedError: Base class raise NotImplementedError
        """
        self.touch(str(Path(path) / ".empty"))

    def rm(self, path: str, recursive: bool = False) -> None:
        """Delete a file

        Args:
            path (str): Path to file to delete
            recursive (bool): If allowed to delete recursive directories or not.

        """
        if recursive:
            raise NotImplementedError("`recursive` not supported")
        self.bucket.Object(path).delete()

    def cp(
        self,
        src_path: str,
        dest_path: str,
        recursive: bool = False,
        overwrite: bool = False,
    ) -> None:
        """Copy a file from source to destination

        Args:
            src_path (str): Path to file or directory to copy.
            dest_path (str): Path to file or directory to copy to.
            recursive (bool): If should copy sub directories as well.
            overwrite (bool): If should copy to destination that already exists.
        """
        if recursive or overwrite:
            raise NotImplementedError("`recursive` and `overwrite` not supported.")

        self.touch(dest_path, self.open(src_path))

    def isdir(self, path: str) -> bool:
        """Check if path is directory or not.

        Args:
            path (str): Path we want to check

        Raises:
            NotImplementedError: Base class raise NotImplementedError

        Returns:
            bool: True/False if path is directory or not
        """
        raise NotImplementedError()

    def isfile(self, path: str) -> bool:
        """Check if path is a file or not.

        Args:
            path (str): Path we want to check

        Raises:
            NotImplementedError: Base class raise NotImplementedError

        Returns:
            bool: True/False if path is file or not.
        """
        raise NotImplementedError()

    def exists(self, path: str) -> bool:
        """Check if path exists or not.

        Args:
            path (str): Path we want to check

        Raises:
            NotImplementedError: Base class raise NotImplementedError

        Returns:
            bool: True/False if path is file or not.
        """
        try:
            self.connection.meta.client.head_object(
                Bucket=default_settings.AWS_STORAGE_BUCKET_NAME, Key=path
            )
            return True
        except ClientError:
            return False



from typing import Any


class BaseDataManager:
    """Base abstract class for data managers
    that define interface.
    """

    def read(self, path: str, root: bool = False, *args, **kwargs) -> Any:
        raise NotImplementedError()

    def read_sql(self, sql: str, *args, **kwargs) -> Any:
        raise NotImplementedError()

    def write(self, path: str, df: Any, root: bool = False, *args, **kwargs) -> None:
        raise NotImplementedError()



from io import BytesIO
import matplotlib.pyplot as plt
from contextlib import contextmanager
from . import default_storage


@contextmanager
def plot(file_path: str, root: bool = False):
    file_path = file_path if root is True else default_storage.runpath(file_path)

    stream = BytesIO()
    try:
        yield plt
    finally:
        plt.savefig(stream)
        plt.close()
        stream.seek(0)
        default_storage.touch(file_path, stream)



# flake8: noqa
from gluepy.conf import default_settings
from gluepy.utils.loading import LazyProxy, import_string, SingletonMixin
from .local import LocalStorage
from .s3 import S3Storage
from .google import GoogleStorage


default_storage = LazyProxy(lambda: import_string(default_settings.STORAGE_BACKEND)())



import os
import logging
from typing import List, Tuple, Union
from io import StringIO, BytesIO
from gluepy.files.storages.base import BaseStorage

logger = logging.getLogger(__name__)


class LocalStorage(BaseStorage):
    """Storage implementation of local file system interactions"""

    MAX_CHUNK_SIZE = 4_000_000

    def touch(self, file_path: str, content: Union[StringIO, BytesIO]) -> None:
        """Create a new blob at file path.

        Args:
            file_path (str): Path to file we want to create
            content (Union[StringIO, BytesIO]): Content of file we want to generate

        Raises:
            TypeError: Raises type error if the chunk read in from the content
              stream is not a valid type.
        """
        # Local file system require directories to exist before blobs are created.
        if not self.exists(os.path.dirname(file_path)):
            self.mkdir(os.path.dirname(file_path))

        logger.debug(f"Writing file to path '{self.abspath(file_path)}'.")
        with open(self.abspath(file_path), mode="wb") as stream:
            while True:
                chunk = content.read(self.MAX_CHUNK_SIZE)
                if not chunk:
                    break

                if isinstance(chunk, str):
                    chunk = chunk.encode("utf-8")
                elif not isinstance(chunk, bytes):
                    raise TypeError("Chunk is neither bytes or string")

                stream.write(chunk)

    def open(self, file_path: str, mode: str = "rb") -> Union[str, bytes]:
        """Opens a blob at file_path

        Args:
            file_path (str): File path of blob we want to open
            mode (str): The read mode of the file, if should return string or bytes.

        Returns:
            bytes: Returns the string or byte content of the file.
        """
        logger.debug(f"Reading file from path '{self.abspath(file_path)}'.")
        with open(self.abspath(file_path), mode=mode) as stream:
            f = stream.read()
        return f

    def ls(self, path: str) -> Tuple[List[str], List[str]]:
        """List all files and directories at given path.

        Args:
            path (str): Path where we want to list contents of

        Raises:
            ValueError: Raises a ValueError if the isdir and isfile functions cannot
                identify if a blob is a file or a dir.

        Returns:
            Tuple[List[str], List[str]]:
              First list is files, second list is directories.
        """
        files = []
        dirs = []
        blobs = [
            self.abspath(os.path.join(path, fpath))
            for fpath in os.listdir(self.abspath(path))
        ]

        for b in blobs:
            if self.isdir(b):
                dirs.append(self.relpath(b))
            elif self.isfile(b):
                files.append(self.relpath(b))
            else:
                raise ValueError(f"File '{b}' is neither a file or directory.")

        return files, dirs

    def mkdir(self, path: str, make_parents: bool = False) -> None:
        """Make a new directory at location

        Args:
            path (str): Path of directory we want to create

        """
        os.makedirs(self.abspath(path), exist_ok=True)

    def isdir(self, path: str) -> bool:
        """Check if path is directory or not.

        Args:
            path (str): Path we want to check

        Raises:
            FileNotFoundError: Raise error if the path cannot be found.

        Returns:
            bool: True/False if path is directory or not
        """
        if not self.exists(path):
            raise FileNotFoundError(f"Path '{path}' does not exist.")

        return os.path.isdir(self.abspath(path))

    def isfile(self, path: str) -> bool:
        """Check if path is a file or not.

        Args:
            path (str): Path we want to check

        Raises:
            FileNotFoundError: Raise error if the path cannot be found.

        Returns:
            bool: True/False if path is file or not.
        """
        if not self.exists(path):
            raise FileNotFoundError(f"Path '{path}' does not exist.")

        return os.path.isfile(self.abspath(path))

    def exists(self, path: str) -> bool:
        """Check if path exists or not.

        Args:
            path (str): Path we want to check

        Returns:
            bool: True/False if path is file or not.
        """
        return os.path.exists(self.abspath(path))
